/* mapview - map viewer for PalmOS
   Copyright (C) 2005  Edwin Steiner <edwin.steiner@gmx.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <PalmOS.h>
#include <string.h>
#include <stdlib.h>
#include "global.h"
#include "rsrc.h"
#include "debug.h"
#include "map.h"
#include "view.h"
#include "maplist.h"
#include "filedlg.h"
#include "mapview.h"

/*** global variables *******************************************************/

static mapview_map_t *g_map = NULL;
static mapview_view_t *g_view = NULL;
static mapview_maplist_t *g_ml = NULL;

UInt32 g_window_manager_version = 0;
bool   g_high_density_feature = false;
UInt32 g_screen_density = 0;
UInt32 g_screen_width = 0;
UInt32 g_screen_height = 0;
UInt32 g_rom_version = 0;

static bool    g_screen_mode_set = false;
static UInt32  g_saved_screen_width;
static UInt32  g_saved_screen_height;
static UInt32  g_saved_screen_depth;
static Boolean g_saved_screen_color;

/*** helper functions *******************************************************/

static void *
mapview_object_from_id(FormPtr form,UInt16 objid)
{
        UInt16 index;
        
        ASSERT(form);

        index = FrmGetObjectIndex(form,objid);
        if (index == frmInvalidObjectId)
                return NULL;

        return FrmGetObjectPtr(form,index);
}

/*** opening / closing maps *************************************************/

static bool
mapview_close(void)
{
        mapview_view_free(g_view);
        g_view = NULL;
        mapview_map_free(g_map);
        g_map = NULL;

        return true;
}

static bool
mapview_open_map_view()
{
        if (!g_map)
                return false;

        if (g_map->density == g_screen_density)
                g_view = mapview_view_new(g_map,kCoordinatesNative,0,0,g_screen_width,g_screen_height);
        else
                g_view = mapview_view_new(g_map,kCoordinatesStandard,0,0,160,160);

        if (!g_view)
                return false;

        mapview_view_move_to_map_center(g_view);
        return true;
}

bool
mapview_open_map(UInt16 cardno,LocalID id)
{
        mapview_close();

        g_map = mapview_map_new(cardno,id);
        if (!g_map)
                return false;

        return mapview_open_map_view();
}

bool
mapview_open_map_mapfile(mapview_mapfile_t *mapfile)
{
        mapview_close();

        g_map = mapview_map_new_mapfile(mapfile);
        if (!g_map)
                return false;

        return mapview_open_map_view();
}

static bool
mapview_open_map_dialog()
{
        bool success;

        if (g_rom_version >= 0x03503000) {
                /* the SiEd file dialog requires PalmOS 3.5+ */
                success = mapview_filedialog_open();
        }
        else {
                mapview_close();
                FrmGotoForm(OpenMapFormID);
                return true;
        }
        FrmGotoForm(MainFormID);
        return success;
}

/*** MainForm ***************************************************************/

static void
MainFormDraw(FormPtr frmP)
{
        if (g_view) {
                mapview_view_draw(g_view);
        }
        else {
                const char *msg = "No map opened.";
                WinDrawChars(msg,strlen(msg),10,25);
        }
}

static bool
in_center_of_screen(Int16 x,Int16 y)
{
        /* XXX always standard coordsys? */
        x -= 80;
        y -= 80;
        return (x > -CENTER_TOLERANCE && x < CENTER_TOLERANCE && 
                y > -CENTER_TOLERANCE && y < CENTER_TOLERANCE);
}

static Boolean 
MainFormHandleEvent(EventPtr e)
{
        WChar key;
        mapview_view_t *view;
        FormPtr frmP = FrmGetActiveForm();
        Int32 x,y;

        view = g_view;

        switch (e->eType) {
                case frmOpenEvent:
                        FrmDrawForm(frmP);
                        MainFormDraw(frmP);
                        return true;

                case frmUpdateEvent:
                        /* XXX FrmDrawForm? */
                        MainFormDraw(frmP);
                        return true;

                case keyDownEvent:
                        if (!view)
                                return false;

                        ASSERT(view);
                      key = e->data.keyDown.chr;
                      switch (key) {
                              case KEY_UP:
                                      mapview_view_move_scaled(view,0,-view->stepsizey);
                                      goto move_tail;
                              case KEY_DOWN:
                                      mapview_view_move_scaled(view,0,+view->stepsizey);
                                      goto move_tail;
                              case KEY_LEFT:
                                      mapview_view_move_scaled(view,-view->stepsizex,0);
                                      goto move_tail;
                              case KEY_RIGHT:
                                      mapview_view_move_scaled(view,+view->stepsizex,0);
move_tail:
                                      MainFormDraw(frmP);
                                      return true;
                              case KEY_MIDDLE:
                              case KEY_SPACE:
                                      if (!mapview_view_down(view))
                                              mapview_view_switch_to_layer(view,0);
                                      MainFormDraw(frmP);
                                      return true;
                              case KEY_BACKSPACE:
                                      if (mapview_view_up(view))
                                              MainFormDraw(frmP);
                                      return true;
                              default:
                                      return false;
                      }
                      return true;

                case penDownEvent:
                      if (!view)
                              return false;

                      if (in_center_of_screen(e->screenX,e->screenY)) {
                              if (!mapview_view_down(view)) {
                                      mapview_view_switch_to_layer(view,0);
                              }
                              MainFormDraw(frmP);
                      }
                      else {
                              mapview_view_screen_to_map(view,e->screenX,e->screenY,&x,&y);
                              mapview_view_move_center(view,x,y);
                              MainFormDraw(frmP);
                      }
                      return true;

                case menuEvent:
                      switch (e->data.menu.itemID) {
                              case MenuOpenMap:
                                      mapview_open_map_dialog();
                                      return true;
                              case MenuAboutMapview:
                                      FrmGotoForm(AboutFormID);
                                      return true;
                      }
                      return true;

                default:
                      return false;
        }
}

/*** OpenMapForm ************************************************************/

static Boolean
OpenMapFormHandleOpen(FormPtr frmP)
{
        ListPtr lst;
        Int16 sel;
        mapview_maplist_item_t *item;

        lst = (ListPtr) mapview_object_from_id(frmP,MapListID);
        if (!lst)
                return true;

        sel = LstGetSelection(lst);
        if (sel == noListSelection)
                return true;

        item = mapview_maplist_nth_item(g_ml,sel);
        if (!item)
                return true;

        if (!mapview_open_map(item->cardno,item->localid))
                return true;
        
        FrmGotoForm(MainFormID);
        return true;
}

static Boolean 
OpenMapFormHandleEvent(EventPtr e)
{
        ListPtr lst;
        FormPtr frmP = FrmGetActiveForm();

        switch (e->eType) {
                case frmOpenEvent:
                        if (!g_ml)
                                g_ml = mapview_maplist_new();
                        if (!g_ml)
                                return true;

                        lst = (ListPtr) mapview_object_from_id(frmP,MapListID);
                        if (!lst)
                                return true;

                        mapview_maplist_set_list_choices(g_ml,lst);

                        FrmDrawForm(frmP);
                        return true;

                case frmUpdateEvent:
                        FrmDrawForm(frmP);
                        return true;

                case ctlSelectEvent:
                        switch (e->data.ctlSelect.controlID) {
                                case ButtonOpenID:
                                        return OpenMapFormHandleOpen(frmP);
                        }
                default:
                        return false;
        }
}

/*** AboutForm **************************************************************/

static Boolean 
AboutFormHandleEvent(EventPtr e)
{
        FormPtr frmP = FrmGetActiveForm();

        switch (e->eType) {
                case frmOpenEvent:
                case frmUpdateEvent:
                        FrmDrawForm(frmP);
                        return true;

                case ctlSelectEvent:
                        switch (e->data.ctlSelect.controlID) {
                                case AboutFormOKButtonID:
                                        FrmGotoForm(MainFormID);
                                        return true;
                        }
                default:
                        return false;
        }
}

/* ApplicationHandleEvent ********************************************

   IN:
       ev...............event to handle

*********************************************************************/

static Boolean 
ApplicationHandleEvent(EventPtr e)
{
        if (e->eType == frmLoadEvent) {
                UInt16 formId = e->data.frmLoad.formID;
                FormPtr frmP = FrmInitForm(formId);
                FrmSetActiveForm(frmP);

                switch(formId) {
                        case MainFormID:
                                FrmSetEventHandler(frmP, MainFormHandleEvent);
                                return true;
                        case AboutFormID:
                                FrmSetEventHandler(frmP, AboutFormHandleEvent);
                                return true;
                        case OpenMapFormID:
                                FrmSetEventHandler(frmP, OpenMapFormHandleEvent);
                                return true;
                }
        }
        return false;
}

/* EventLoop *********************************************************

   main event loop

*********************************************************************/

static void 
EventLoop(void) 
{ 
        UInt16 error; 
        EventType event; 
        do { 
                EvtGetEvent(&event, evtWaitForever); 

                if (!SysHandleEvent(&event)) { 
                        if (!MenuHandleEvent(NULL, &event, &error)) {
                                if (!ApplicationHandleEvent(&event)) {
                                        FrmDispatchEvent(&event);
                                }
                        }
                }
        } while (event.eType != appStopEvent); 
} 

/* InitMapview *******************************************************

   initialize global data, check features

*********************************************************************/

static void 
InitMapview(void) 
{ 
        Err err;

        /* request rom version */
        err = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &g_rom_version); 
        if (err != 0)
                g_rom_version = 0;

        /* request window manager version */
        err = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &g_window_manager_version);
        if (err != 0)
                g_window_manager_version = 0;

        /* check for high density feature set */
        g_high_density_feature = (g_window_manager_version >= 4);

        /* get screen attributes */
        if (g_high_density_feature) {
                err = WinScreenGetAttribute(winScreenDensity, &g_screen_density); 
                if (err != errNone)
                        goto default_screen;
                err = WinScreenGetAttribute(winScreenWidth, &g_screen_width);
                if (err != errNone)
                        goto default_screen;
                err = WinScreenGetAttribute(winScreenHeight, &g_screen_height);
                if (err != errNone)
                        goto default_screen;
        }
        else {
default_screen:
                /* Fall back to a low-density Palm screen */
                g_high_density_feature = false;
                g_screen_density = 72;
                g_screen_width = 160;
                g_screen_height = 160;
        }

        mapview_filedialog_init();
} 

/* TryScreenMode *****************************************************

   try to set the screen mode with the given depth

Returns:
   true...the mode could be set. In this case g_screen_mode_set is
          set to true if the screen mode has been changed.
   false..otherwise (g_screen_mode_set is not touched).

*********************************************************************/

static bool
TryScreenMode(UInt32 depth)
{
        Err err;

        /* if the screen is already set to the given mode */
        /* there is nothing to do                         */
        if (depth == g_saved_screen_depth)
                return true;

        err = WinScreenMode(winScreenModeSet,&g_saved_screen_width,
                                             &g_saved_screen_height,
                                             &depth,
                                             NULL);

        if (err == errNone) {
                g_screen_mode_set = true;
                return true;
        }
        else {
                return false;
        }
}

/* InitScreen ********************************************************

   set the screen to the best mode we can get. If the screen mode is
   changed, g_screen_mode_set is set to true.

   XXX It would be best to match the screen mode to the map bitmaps,
   but right now the bit depth/color is not stored in the maps. Maybe
   this should be introduced in version 0.2.0.

*********************************************************************/

static void
InitScreen(void)
{
        Err err;
        UInt32 monodepths;
        UInt32 colordepths;
        UInt32 anydepths;
        Boolean color;

        /* the following requires the 3.5 feature set */
        if (g_rom_version < 0x03503000)
                return;

        /* save the current screen mode */
        err = WinScreenMode(winScreenModeGet,&g_saved_screen_width,
                                             &g_saved_screen_height,
                                             &g_saved_screen_depth,
                                             &g_saved_screen_color);
        if (err != errNone)
                return;

        /* request which depths are supported in mono/color mode  */
        /* (All devices I had available return the same value for */
        /*  mono and color, but who knows...)                     */
        color = false;
        err = WinScreenMode(winScreenModeGetSupportedDepths,NULL,NULL,&monodepths,&color);
        if (err != errNone)
                return;
        color = true;
        err = WinScreenMode(winScreenModeGetSupportedDepths,NULL,NULL,&colordepths,&color);
        if (err != errNone)
                return;
        anydepths = monodepths | colordepths;

        /* if the highest depth is already set, we do not change anything */
        if ((anydepths >> g_saved_screen_depth) == 0)
                return;

        /* try to set 8-bit color mode */
        if ((colordepths & (1 << 7)) != 0) {
                if (TryScreenMode(8))
                        return;
        }

        /* try to set 4-bit greyscale mode */
        if ((monodepths & (1 << 3)) != 0) {
                if (TryScreenMode(4))
                        return;
        }

        /* try to set 2-bit greyscale mode */
        if ((monodepths & (1 << 1)) != 0) {
                if (TryScreenMode(2))
                        return;
        }

        /* leave the default, which is probably black&white */
}

/* ResetScreen *******************************************************

   restore the screen mode that was active before InitScreen

*********************************************************************/

static void
ResetScreen(void)
{
        Err err;

        /* no need to reset it if we did not change it */
        if (!g_screen_mode_set)
                return;

        err = WinScreenMode(winScreenModeSet,&g_saved_screen_width,
                                             &g_saved_screen_height,
                                             &g_saved_screen_depth,
                                             &g_saved_screen_color);
        /* no meaningful reaction to errors possible */
}

/* PilotMain *********************************************************

   main entry point

*********************************************************************/

UInt32 
PilotMain(UInt16 cmd,void *cmdPBP,UInt16 launchFlags)
{
        if (cmd == sysAppLaunchCmdNormalLaunch) {
                InitMapview();
                InitScreen();

                mapview_open_map_dialog();

                EventLoop();

                FrmCloseAllForms();

                ResetScreen();

                mapview_maplist_free(g_ml);
                mapview_view_free(g_view);
                mapview_map_free(g_map);
                mapview_filedialog_cleanup();
        }
        return 0;
}

/* vim: et sts=8 sw=8 ai
 */

